# MCSLA Core Library

The MCSLA Core Library provides 

## Table of Contents
1. [Features](#features)
1. [Installation](#installation)
1. [Getting started](#getting-started)
1. [The Interface](#the-interface)
1. [License](#license)

## Features

 * Support for different aggregation expressions
 * Aggregation of cloud service SLAs
 * Evaluation of a single cloud service SLA metric
 * Evaluation of aggregated cloud services SLA metrics

## Building from Source

The project is available via Git repository. If you have access, do the following steps:

```shell
$> git clone https://gitlab.fokus.fraunhofer.de/DECIDE/mcsla-core.git
$> cd mcsla-core
```

The project uses Maven as build tool. After build you will find the jar and a fat jar in the target directory. To build use the follwoing command:

```shell
$> mvn clean package
```

## Installation

For non-Maven based projects you can take the build jar file located in the target directory after executing the build command and put it in the classpath of your application. There is also a fat jar provided containing all dependencies if required.

For Maven based projects you need to install it in a Maven repository which your application can access. E.g. to put it in your local maven repository, you can simply call
`$> mvn install`

Finally, your application pom.xml requires the following dependency:

```xml
<dependency>
    <groupId>eu.DECIDEh2020</groupId>
    <artifactId>mcsla-core</artifactId>
    <version>0.0.1-SNAPSHOT</version>
</dependency>
```

## Getting started
A small example how to create a metrics context and evaluate a cloud service objective:  

```java
// create context
MetricsContext context = MetricsContext.create(appDescription);

// get sla service objective metrics for a apecific cloud service
Map<String, ServiceObjectiveMetrics> metrics = context.getCsMetrics(csId);

// get the service objectove metrics for a specific term name, e.g. "Availability"
ServiceObjeciveMetrics availabilityMetrics = metrics.get("Availability");

// now evaluate against a concrete monitoring result
EvaluationResult<Double> evaluationResult = availabilityMetrics.evaluate(monitoringResult);

if (!evaluationResult.isError()) {
    log.info("condition met: {}, result: {}", evaluationResult.isConditionMet(), evaluationResult.getMeasurementResultValue());
} else {
    log.error(evaluationResult.getErrorMessage());
}
```

For further examples please take a look at the test cases.

## The Interface

In general there are three main classes:
 * [MetricsContext](#metricscontext)
 * [ServiceObjectiveMetrics](#serviceobjectivemetrics)
 * [EvaluationResult\<T\>](#evaluationresultt)

### MetricsContext
This class provides a static methods for creating context objects. Created context can be used to get service objective metrics of 
each SLA as a map. Furthermore, it contains a memory for cloud service evaluation results, so specific aggregation
expressions can be implicitly applied. And finally, it allows the aggregation of service objectives or raw values.

#### Aggregation
The easiest way to use the MetricsContext is to aggregate some raw values:
```
List<Double> values = Arrays.asList(99.9, 99.8, 99.95);
Double value = MetricsContext.aggregate(values, Predefined.AVAILABILITY_AGGREGATION_SUMTYPE);
```
See [Predefined](#predefined-aggregation-expressions) for available aggregation expressions.

To start working with the library in the context of an application description you need to create a MetricsContext object:
```java
AppDescription appDescription = ...
MetricsContext context = MetricsContext.create(appDescription);
```
You can now aggregate all agreement values of all cloud services for a specific term name: 
```java
Double value = MetricsContext.aggregate(Predefined.AVAILABILITY_AGGREGATION_SUMTYPE, "Availability");
```
The return type depends on the predefined aggregation expression to be used.

#### Evaluation
The MetricsContext object has two methods for retrieving [ServiceObjectiveMetrics](#serviceobjectivemetrics) which are needed for evaluation:
```java
Map<String, ServiceObjectiveMetrics> appMetrics = context.getAppMetrics();
Map<String, ServiceObjectiveMetrics> cloudserviceMetrics = context.getCsMetrics(csId);
```
The return value is a map of term names and corresponding metrics of either a specific cloud service or the application.
Now you can get the [ServiceObjectiveMetrics](#serviceobjectivemetrics) object for a specific term name:
```java
ServiceObjectiveMetrics objectiveMetrics = appMetrics.get("Availability");
```

### ServiceObjectiveMetrics
You need objects s of this class to evaluate the metrics. The class has the following evaluation functions:
* evaluate()
* evaluate(double[] measuredValues)
* evaluate(int[] measuredValues)
* evaluate(List\<Integer\> measuredValues)
* evaluate(Map\<String, List\<Long\>\> measuredValues)
* evaluate(double measuredValue)

They all return a [EvaluationResult\<T\>](#evaluationresultt) object.

The first one without passing any monitoring results utilizes the memory function of the metrics context. It therefore requires that each monitoring result to be aggregate must be evaluated beforehand.
This makes sense where you have to evaluate first each cloud service metrics and then the aggregated application metric:
```java
Map<String, ServiceObjectiveMetrics> cloudserviceOneMetrics = context.getCsMetrics("1");
ServiceObjectiveMetrics oneMetrics = cloudserviceOneMetrics.get("Availability");
EvaluationResult<Double> resultOne = OneMetrics.evaluate(99.9);
...
Map<String, ServiceObjectiveMetrics> cloudserviceTwoMetrics = context.getCsMetrics("1");
ServiceObjectiveMetrics twoMetrics = cloudserviceTwoMetrics.get("Availability");
EvaluationResult<Double> resultTwo = twoMetrics.evaluate(99.9);
...
Map<String, ServiceObjectiveMetrics> appMetrics = context.getAppMetrics();
ServiceObjectiveMetrics availabilityMetrics = appMetrics.get("Availability");
EvaluationResult<Double> resultApp = availabilityMetrics.evaluate();
```

### EvaluationResult\<T\>
The first thing you should do is to check if the evaluation was successful:
```java
EvaluationResult<Doube> result = ...
if (result.isError()) {
    log.error(result.getErrorMesssge());
} else {
    ...
}
```
A successful evaluation result contains the following information:
* The evaluation date and time
* The underlying term name
* The measurement result value
* The used condition statement
* And finally if the condition is met (see [Statement](#condition-statements))

```java
EvaluationResult<T> result = ...
Date time = result.getMeasurementTime();
String termName = result.getTermName();
T measurementValue = result.getMeasurementResultValue();
Statement conditionStatement ? result.getConditionStatement();
assert result.isConditionMet() : "Condition is not met.";
```

### Predefined Aggregation Expressions
| Expression                       |                                                                                     |
|----------------------------------|-------------------------------------------------------------------------------------|
| EMPTY                            | The empty aggregation returns always null.                                          |
| AVAILABILITY_UPTIME_BC           | Evaluation of downtimes in seconds to an percentage uptime during a billing cycle.  |
| AVAILABILITY_MTBFMTTR            | Evaluation based on meantime between failure and meantime between recovery.         |
| AVAILABILITY_AGGREGATION_SUMTYPE | An aggregation that summarizes percentage values                                    |
| AGGREGATION_MINTYPE              | Returns the min value.                                                              |
| AGGREGATION_MAXTYPE              | Returns the max value                                                               |
| AGGREGATION_SUMTYPE              | Returns the sum up of values.                                                       |

### Condition Statements
* greater
* greaterOrEqual
* less
* lessOrEqual
* equal

## License

[GNU Affero General Public License Version 3](LICENSE.txt)