/*
 * Copyright (c) 2017 Fraunhofer FOKUS.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the
 * GNU AGPL v3 which accompanies
 * this distribution, and is available at
 * http://www.gnu.org/licenses/agpl.txt
 *
 * Contributors:
 * Simon Dutkowski Fraunhofer FOKUS
 *
 * Initially developed in the context of DECIDE EU project www.DECIDE-h2020.eu
 */
package eu.DECIDEh2020.mcsla.service.repositories;

import eu.DECIDEh2020.appManager.AppManager;
import eu.DECIDEh2020.appManager.exceptions.AppManagerException;
import eu.DECIDEh2020.appManager.exceptions.DECIDEValidationException;
import eu.DECIDEh2020.appManager.models.*;
import eu.DECIDEh2020.mcsla.core.metrics.MetricsContext;
import eu.DECIDEh2020.mcsla.core.metrics.Predefined;
import io.vertx.core.AsyncResult;
import io.vertx.core.Future;
import io.vertx.core.Handler;
import io.vertx.core.buffer.Buffer;
import io.vertx.core.json.JsonObject;
import io.vertx.ext.web.client.HttpRequest;
import org.everit.json.schema.ValidationException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.io.IOException;
import java.nio.file.Path;
import java.util.*;
import java.util.stream.Collectors;

public class RepositoriesServiceImpl implements RepositoriesService {

    private final Logger log = LoggerFactory.getLogger(getClass());

    private Map<String, AppManager> local = new HashMap<>();

    private Path base;
    private HttpRequest<Buffer> acsmiRequest;

    RepositoriesServiceImpl(Path base, HttpRequest<Buffer> acsmiRequest, Handler<AsyncResult<RepositoriesService>> readyHandler) {
        this.base = base;
        this.acsmiRequest = acsmiRequest;
        readyHandler.handle(Future.succeededFuture(this));
    }

    @Override
    public RepositoriesService isLocal(String appName, Handler<AsyncResult<Boolean>> handler) {
        handler.handle(Future.succeededFuture(local.containsKey(appName)));
        return this;
    }

    @Override
    public RepositoriesService initRepository(String ref, String token, String username, String password, Handler<AsyncResult<JsonObject>> handler) {
        log.debug("initialize {} application", ref);
        Path repo = base.resolve(ref.replaceAll("[^\\w\\s]", ""));

        try {
            final AppManager appManager;
            if (token != null && !token.isEmpty()) {
                appManager = AppManager.open(ref, token, repo);
            } else {
                appManager = AppManager.open(ref, username, password, repo);
            }
            appManager.sync();

            AppDescription appDescription = appManager.getAppDescription();
            String name = appDescription.getName().toLowerCase().replace(" ", "_");
            local.put(name, appManager);

            if (appDescription.getMcsla() == null || appDescription.getMcsla().getSla() == null) {
                addNfrsAsSla(appDescription);
                appManager.write(appDescription);
            }

            fetchSlas(appDescription, res -> {
                if (res.succeeded()) {
                    try {
                        appManager.write(appDescription);
                        handler.handle(Future.succeededFuture(new JsonObject().put("name", name).put("app", JsonObject.mapFrom(appDescription))));
                    } catch (AppManagerException | IOException e) {
                        log.error("write app description", e);
                        Throwable cause = res.cause();
                        if (cause instanceof DECIDEValidationException) {
                            ValidationException coreException = (ValidationException)cause.getCause();
                            String message = coreException.getAllMessages().toString();
                            handler.handle(Future.failedFuture(message));
                        } else {
                            handler.handle(Future.failedFuture(e));
                        }
                    }
                } else {
                    handler.handle(Future.failedFuture(res.cause()));
                }
            });

        } catch (AppManagerException | IOException e) {
            log.error("reading app description", e);
            if (e instanceof DECIDEValidationException) {
                ValidationException coreException = (ValidationException)e.getCause();
                String message = coreException.getAllMessages().toString();
                handler.handle(Future.failedFuture(message));
            } else {
                handler.handle(Future.failedFuture(e));
            }
        }

        return this;
    }

    @Override
    public RepositoriesService getApplication(String appName, Handler<AsyncResult<JsonObject>> handler) {
        AppManager appManager = local.get(appName);
        if (appManager != null) {
            try {
                handler.handle(Future.succeededFuture(JsonObject.mapFrom(appManager.getAppDescription())));
            } catch (DECIDEValidationException | IOException e) {
                log.error("reading app description", e);
                ValidationException coreException = (ValidationException)e.getCause();
                String message = coreException.getAllMessages().toString();
                handler.handle(Future.failedFuture(message));
            }
        } else {
            log.error("No repo for application name {} found.", appName);
            handler.handle(Future.failedFuture("No repo for application name " + appName + " found."));
        }
        return this;
    }

    @Override
    public RepositoriesService resetApplication(String appName, Handler<AsyncResult<JsonObject>> handler) {
        AppManager appManager = local.get(appName);
        if (appManager != null) {
            try {
                appManager.reset();
                AppDescription appDescription = appManager.getAppDescription();
                if (appDescription.getMcsla() == null || appDescription.getMcsla().getSla() == null) {
                    addNfrsAsSla(appDescription);
                    appManager.write(appDescription);
                }

                fetchSlas(appDescription, res -> {
                    if (res.succeeded()) {
                        try {
                            appManager.write(appDescription);
                            handler.handle(Future.succeededFuture(JsonObject.mapFrom(appManager.getAppDescription())));
                        } catch (AppManagerException | IOException e) {
                            log.error("write app description", e);
                            Throwable cause = res.cause();
                            if (cause instanceof DECIDEValidationException) {
                                ValidationException coreException = (ValidationException)cause.getCause();
                                String message = coreException.getAllMessages().toString();
                                handler.handle(Future.failedFuture(message));
                            } else {
                                handler.handle(Future.failedFuture(e));
                            }
                        }
                    } else {
                        handler.handle(Future.failedFuture(res.cause()));
                    }
                });

            } catch (DECIDEValidationException e) {
                log.error("reading app description", e);
                ValidationException coreException = (ValidationException)e.getCause();
                String message = coreException.getAllMessages().toString();
                handler.handle(Future.failedFuture(message));
            } catch (AppManagerException | IOException e) {
                handler.handle(Future.failedFuture(e));
            }
        } else {
            log.error("No repo for application name {} found.", appName);
            handler.handle(Future.failedFuture("No repo for application name " + appName + " found."));
        }
        return this;
    }

    @Override
    public RepositoriesService updateMcsla(String appName, JsonObject mcsla, Handler<AsyncResult<JsonObject>> handler) {
        AppManager appManager = local.get(appName);
        if (appManager != null) {
            try {
                AppDescription appDescription = appManager.getAppDescription();
                appDescription.setMcsla(mcsla.mapTo(Mcsla.class));
                appManager.writeAndSync(appDescription, "mcsla updated via mcsla service");
                handler.handle(Future.succeededFuture(new JsonObject().put("success", "true")));
            } catch (AppManagerException | IOException e) {
                log.error("Sync with remote", e);
                if (e instanceof DECIDEValidationException) {
                    ValidationException coreException = (ValidationException)e.getCause();
                    String message = coreException.getAllMessages().toString();
                    handler.handle(Future.failedFuture(message));
                } else {
                    handler.handle(Future.failedFuture(e));
                }
            }
        } else {
            log.error("no repo for application name {} found.", appName);
            handler.handle(Future.failedFuture("No repo for application name " + appName + " found."));
        }

        return this;
    }

    private void fetchSlas(AppDescription appDescription, Handler<AsyncResult<Void>> handler) {
        List<SchemaElement> schema = appDescription.getSchema();
        if (schema != null && !schema.isEmpty()) {
            List<String> csIds = schema.stream().map(SchemaElement::getCsId).collect(Collectors.toList());

            if (acsmiRequest != null) {
                acsmiRequest.setQueryParam("serviceids", String.join(",", csIds)).send(res ->{
                    if (res.succeeded()) {
                        if (res.result().statusCode() == 200) {
                            Map<String, Sla> slas = SlaImport.fromResult(res.result().bodyAsJsonArray());
                            appDescription.getMcsla().getCsSlas().putAll(slas);
                            handler.handle(Future.succeededFuture());
                        } else {
                            handler.handle(Future.failedFuture("Service discovery response: " + res.result().statusCode() + " - " + res.result().statusMessage()));
                        }
                    } else {
                        handler.handle(Future.failedFuture(res.cause()));
                    }
                });
            } else {
                Map<String, Sla> slas = SlaImport.fromResource(csIds);
                appDescription.getMcsla().getCsSlas().putAll(slas);
                handler.handle(Future.succeededFuture());
            }
        } else {
            handler.handle(Future.succeededFuture());
        }
    }

    private void addNfrsAsSla(AppDescription appDescription) {
        List<Nfr> applicationNfrs = appDescription.getNfrs().stream().filter(nfr -> nfr.getTags().contains("Application")).collect(Collectors.toList());
        List<Microservice> microservices = appDescription.getMicroservices();
        Sla sla = SlaFactory.getInstance().createFromNfrs(applicationNfrs, microservices);
        sla.setDescription(appDescription.getName() + " Service Level Agreement");
        if (appDescription.getMcsla() == null) {
            appDescription.setMcsla(new Mcsla());
        }
        appDescription.getMcsla().setSla(sla);
    }

    @Override
    public RepositoriesService aggregateServiceObjectives(String appName, String termName, String predefined, Handler<AsyncResult<JsonObject>> handler) {
        AppManager appManager = local.get(appName);
        if (appManager != null) {
            try {
                MetricsContext context = MetricsContext.create(appManager.getAppDescription());
                Object result = context.aggregate(Predefined.valueOf(predefined), termName);

                handler.handle(Future.succeededFuture(JsonObject.mapFrom(result)));

            } catch (DECIDEValidationException | IOException e) {
                log.error("reading app description", e);
                ValidationException coreException = (ValidationException)e.getCause();
                String message = coreException.getAllMessages().toString();
                handler.handle(Future.failedFuture(message));
            }
        } else {
            log.error("No repo for application name {} found.", appName);
            handler.handle(Future.failedFuture("No repo for application name " + appName + " found."));
        }
        return this;
    }

}
