import React, { Component } from "react";
import { connect } from 'react-redux';
import { withRouter } from 'react-router-dom';
import { withStyles } from '@material-ui/core/styles';
import Button from '@material-ui/core/Button';
import Typography from "@material-ui/core/Typography";
import ServiceObjective from './ServiceObjective';
import TextField from '@material-ui/core/TextField';
import InputAdornment from '@material-ui/core/InputAdornment';
import FormGroup from '@material-ui/core/FormGroup';
import FormControl from '@material-ui/core/FormControl';
import FormControlLabel from '@material-ui/core/FormControlLabel';
import Checkbox from '@material-ui/core/Checkbox';
import VisibilityIcon from '@material-ui/icons/Visibility';
import VisibilityOffIcon from '@material-ui/icons/VisibilityOff';
import Grid from "@material-ui/core/Grid";
import Divider from "@material-ui/core/Divider";
import { setSlaDescription, setSlaValidity, setSlaVisibility, addCoveredService, addObjective, deleteObjective } from "../actions/editmcsla";
import { updateMcslaRequest } from '../actions/mcsla';
import { resetError } from '../actions/app';
import { resetRepositoryRequest } from '../actions/repository';
import CoveredServices from "./CoveredServices";
import SpeedDial from '@material-ui/lab/SpeedDial';
import SpeedDialAction from '@material-ui/lab/SpeedDialAction';
import SpeedDialIcon from '@material-ui/lab/SpeedDialIcon';
import UpdateIcon from '@material-ui/icons/Update';
import NetworkCheckIcon from '@material-ui/icons/NetworkCheck';
import LocationIcon from '@material-ui/icons/LocationOn';
import EuroSymbolIcon from '@material-ui/icons/EuroSymbol';
import BlurOnIcon from '@material-ui/icons/BlurOn';
import Snackbar from '@material-ui/core/Snackbar';
import IconButton from '@material-ui/core/IconButton';
import CloseIcon from '@material-ui/icons/Close';

const styles = theme => ({
    flex: {
        flex: 1
    },
    row: {
        display: 'flex',
        justifyContent: 'center',
    },
      button: {
        margin: theme.spacing.unit,
      },
      spacing: {
        marginLeft: theme.spacing.unit,
      },
      fab: {
        position: 'fixed',
        bottom: theme.spacing.unit * 2,
        right: theme.spacing.unit * 2,
      }
});

const speedDialActions = [
    {
        key: "Availability",
        tooltip: "Availability",
        icon: <UpdateIcon/>
    },
    {
        key: "Performance",
        tooltip: "Performance",
        icon: <NetworkCheckIcon/>
    },
    {
        key: "Scalability",
        tooltip: "Scalability",
        icon: <BlurOnIcon/>
    },
    {
        key: "Location",
        tooltip: "Location",
        icon: <LocationIcon/>
    },
    {
        key: "Cost",
        tooltip: "Cost",
        icon: <EuroSymbolIcon/>
    }
]

class McslaView extends Component {
    state = {
        open: false
    }

    handleVisibility(sla) {
        this.props.setVisibility(!sla.visibility);
    }

    handleChangeDescription(event) {
        this.props.setDescription(event.target.value);
    }

    handleChangeValidity(event) {
        this.props.setValidity(event.target.value);
    }

    handleOpen() {
        this.setState({
            open: !this.state.open,
        });
    }

    handleCommit() {
        this.props.updateMcsla(this.props.appName, { sla: this.props.sla, csSlas: this.props.csSlas });
    }

    handleReset() {
        this.props.resetRepo();
    }

    handleErrorClose() {
        this.props.resetError();
    }

    handleAdd(termName) {
        this.props.addObjective(termName);
        this.setState({
            open: !this.state.open,
        });
    }

    handleDelete(termName) {
        this.props.deleteObjective(termName);
    }

    render() {
        const { classes, sla, csSlas } = this.props;
        const { open } = this.state;

        return (
            <div>
                <Grid container spacing={24}>
                    <Grid item md={12}>
                        <div className={classes.row}>
                            <Typography color="textSecondary" variant="headline" paragraph={true}>Application Multi-Cloud SLA</Typography>
                        </div>
                    </Grid>
                    <Grid item md={6}>
                        <TextField fullWidth multiline rowsMax="4" label="Description" value={sla.description} margin="normal" onChange={event => this.handleChangeDescription(event)} />
                        <CoveredServices coveredServices={sla.coveredServices} />
                    </Grid>
                    <Grid item md={2} />
                    <Grid item md={3}>
                        <FormGroup >
                            <FormControl margin="normal">
                                <FormControlLabel label="Visibility" control={<Checkbox icon={<VisibilityOffIcon />} checkedIcon={<VisibilityIcon />} color="primary" checked={sla.visibility} onChange={() => this.handleVisibility(sla)}/>} />
                            </FormControl>
                            <TextField type="number" label="Validity period" value={sla.validityPeriod} margin="normal" InputProps={{inputProps: {min: 0}, startAdornment: <InputAdornment position="start" >Days</InputAdornment>}} onChange={event => this.handleChangeValidity(event)} />
                        </FormGroup>
                    </Grid>
                    <Grid item md={12}>
                        {sla.objectives ? 
                            Object.values(sla.objectives).map(objective => {
                                return (
                                    <ServiceObjective key={objective.termName} objective={objective} onClick={() => this.handleDelete(objective.termName)} /> 
                                )
                            })
                            :
                            <Typography color="textSecondary" variant="subheading" paragraph={true}>No Service Objectives defined.</Typography>
                        }
                    </Grid>
                    <Grid item md={12}>
                        <FormGroup row>
                            <FormControl margin="normal">
                                <Button variant="outlined" color="primary" className={classes.button} onClick={() => this.handleCommit()}>Commit</Button>
                            </FormControl>
                            <FormControl margin="normal">
                                <Button variant="outlined" color="secondary" className={classes.button} onClick={() => this.handleReset()}>Reset</Button>
                            </FormControl>
                        </FormGroup>
                    </Grid>
                    <Grid item md={12}>
                        <Divider/><br/>
                        <div className={classes.row}>
                            <Typography color="textSecondary" variant="headline" paragraph={true}>Cloud Services SLAs</Typography>
                        </div>
                    </Grid>
                    <Grid item md={12}>
                        {Object.keys(csSlas).map(csSla => {
                            return (
                                <div key={csSla}>
                                    <FormGroup row>
                                        {csSla.visibility ? <VisibilityIcon/> : <VisibilityOffIcon/>}
                                        <Typography className={classes.spacing} variant="subheading">{csSlas[csSla].description}</Typography>
                                    </FormGroup>
                                    <CoveredServices readOnly={true} coveredServices={csSlas[csSla].coveredServices} />
                                    <br/>
                                    {csSlas[csSla].objectives ? Object.values(csSlas[csSla].objectives).map(objective => {
                                        return (
                                            <ServiceObjective key={objective.termName} objective={objective} readonly />
                                        )
                                    })
                                    :
                                    <Typography color="textSecondary" variant="subheading" paragraph={true}>No Service Objectives defined.</Typography>
                                    }
                                    <br/>
                                    <br/>
                                </div>
                            )
                        })}
                    </Grid>
                </Grid>
                <Snackbar
                    anchorOrigin={{
                        vertical: 'bottom',
                        horizontal: 'left',
                    }}
                    open={this.props.error.length > 0}
                    message={<span>{this.props.error}</span>} 
                    action={[
                        <IconButton key="close" aria-label="Close" color="inherit"
                            onClick={() => this.handleErrorClose()}>
                            <CloseIcon />
                        </IconButton>
                    ]}/>
                <SpeedDial className={classes.fab} icon={<SpeedDialIcon/>} ariaLabel="Add Service Objective" open={open} onClick={() => this.handleOpen()}>
                    {Object.values(this.props.speedDialActions).map(speedDial => {
                        return (
                            <SpeedDialAction
                                key={speedDial.key}
                                icon={speedDial.icon}
                                tooltipTitle={speedDial.tooltip}
                                onClick={() => this.handleAdd(speedDial.key)} />
                        )                        
                    })}
                </SpeedDial>
            </div>
        );
    }
}

const mapStateToProps = state => {
    const sda = speedDialActions.filter(s => {
        return state.decideApp.mcsla.sla.objectives ? state.decideApp.mcsla.sla.objectives.find(o => o.termName === s.key) === undefined : true;
    });
    return {
        appName: state.app.appName,
        sla: state.decideApp.mcsla.sla,
        csSlas: state.decideApp.mcsla.csSlas,
        speedDialActions: sda,
        error: state.app.error.message
    }
}

const mapDispatchToProps = dispatch => {
    return {
        setDescription: description => dispatch(setSlaDescription(description)),
        setValidity: validity => dispatch(setSlaValidity(validity)),
        setVisibility: visibility => dispatch(setSlaVisibility(visibility)),
        addService: service => dispatch(addCoveredService(service)),
        updateMcsla: (name, mcsla) => dispatch(updateMcslaRequest(name, mcsla)),
        deleteObjective: termName => dispatch(deleteObjective(termName)),
        addObjective: termName => dispatch(addObjective(termName)),
        resetError: () => dispatch(resetError()),
        resetRepo: () => dispatch(resetRepositoryRequest())
    }
}

export default withRouter(connect(mapStateToProps, mapDispatchToProps)(withStyles(styles, { withTheme: true })(McslaView)))
